@@ -1,31 +1,42 @@
-use ctype::{isalpha, isdigit, ispunct, isspace};
-use nodes::TableAlignment;
-use nodes::{
-    AstNode, ListDelimType, ListType, NodeCodeBlock, NodeHeading, NodeHtmlBlock, NodeLink,
-    NodeValue,
+use crate::ctype::{isalpha, isdigit, ispunct, isspace};
+use crate::nodes::{
+    AstNode, ListDelimType, ListType, NodeAlert, NodeCodeBlock, NodeHeading, NodeHtmlBlock,
+    NodeLink, NodeMath, NodeTable, NodeValue, NodeWikiLink,
 };
-use parser::ComrakOptions;
-use scanners;
-use std;
+use crate::nodes::{NodeList, TableAlignment};
+#[cfg(feature = "shortcodes")]
+use crate::parser::shortcodes::NodeShortCode;
+use crate::parser::{Options, WikiLinksMode};
+use crate::scanners;
+use crate::strings::trim_start_match;
+use crate::{nodes, Plugins};
+
 use std::cmp::max;
 use std::io::{self, Write};
-use {nodes, ComrakPlugins};
 
 /// Formats an AST as CommonMark, modified by the given options.
 pub fn format_document<'a>(
     root: &'a AstNode<'a>,
-    options: &ComrakOptions,
+    options: &Options,
     output: &mut dyn Write,
 ) -> io::Result<()> {
-    format_document_with_plugins(root, options, output, &ComrakPlugins::default())
+    // Formatting an ill-formed AST might lead to invalid output. However, we don't want to pay for
+    // validation in normal workflow. As a middleground, we validate the AST in debug builds. See
+    // https://github.com/kivikakk/comrak/issues/371.
+    #[cfg(debug_assertions)]
+    root.validate().unwrap_or_else(|e| {
+        panic!("The document to format is ill-formed: {:?}", e);
+    });
+
+    format_document_with_plugins(root, options, output, &Plugins::default())
 }
 
 /// Formats an AST as CommonMark, modified by the given options. Accepts custom plugins.
 pub fn format_document_with_plugins<'a>(
     root: &'a AstNode<'a>,
-    options: &ComrakOptions,
+    options: &Options,
     output: &mut dyn Write,
-    _plugins: &ComrakPlugins,
+    _plugins: &Plugins,
 ) -> io::Result<()> {
     let mut f = CommonMarkFormatter::new(root, options);
     f.format(root);
@@ -36,9 +47,9 @@ pub fn format_document_with_plugins<'a>(
     Ok(())
 }
 
-struct CommonMarkFormatter<'a, 'o> {
+struct CommonMarkFormatter<'a, 'o, 'c> {
     node: &'a AstNode<'a>,
-    options: &'o ComrakOptions,
+    options: &'o Options<'c>,
     v: Vec<u8>,
     prefix: Vec<u8>,
     column: usize,
@@ -50,17 +61,18 @@ struct CommonMarkFormatter<'a, 'o> {
     in_tight_list_item: bool,
     custom_escape: Option<fn(&'a AstNode<'a>, u8) -> bool>,
     footnote_ix: u32,
+    ol_stack: Vec<usize>,
 }
 
 #[derive(PartialEq, Clone, Copy)]
 enum Escaping {
     Literal,
     Normal,
-    URL,
+    Url,
     Title,
 }
 
-impl<'a, 'o> Write for CommonMarkFormatter<'a, 'o> {
+impl<'a, 'o, 'c> Write for CommonMarkFormatter<'a, 'o, 'c> {
     fn write(&mut self, buf: &[u8]) -> std::io::Result<usize> {
         self.output(buf, false, Escaping::Literal);
         Ok(buf.len())
@@ -71,8 +83,8 @@ impl<'a, 'o> Write for CommonMarkFormatter<'a, 'o> {
     }
 }
 
-impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
-    fn new(node: &'a AstNode<'a>, options: &'o ComrakOptions) -> Self {
+impl<'a, 'o, 'c> CommonMarkFormatter<'a, 'o, 'c> {
+    fn new(node: &'a AstNode<'a>, options: &'o Options<'c>) -> Self {
         CommonMarkFormatter {
             node,
             options,
@@ -87,6 +99,7 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
             in_tight_list_item: false,
             custom_escape: None,
             footnote_ix: 0,
+            ol_stack: vec![],
         }
     }
 
@@ -108,6 +121,7 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
                 }
             }
             self.column = 0;
+            self.last_breakable = 0;
             self.begin_line = true;
             self.begin_content = true;
             self.need_cr -= 1;
@@ -206,7 +220,7 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
                         && (c == b'.' || c == b')')
                         && follows_digit
                         && (nextc == 0 || isspace(nextc)))))
-                || (escaping == Escaping::URL
+                || (escaping == Escaping::Url
                     && (c == b'`'
                         || c == b'<'
                         || c == b'>'
@@ -218,7 +232,7 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
                     && (c == b'`' || c == b'<' || c == b'>' || c == b'"' || c == b'\\')));
 
         if needs_escaping {
-            if escaping == Escaping::URL && isspace(c) {
+            if escaping == Escaping::Url && isspace(c) {
                 write!(self.v, "%{:2X}", c).unwrap();
                 self.column += 3;
             } else if ispunct(c) {
@@ -273,11 +287,14 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
             None => return false,
         };
 
-        if let NodeValue::Item(..) = tmp.data.borrow().value {
-            if let NodeValue::List(ref nl) = tmp.parent().unwrap().data.borrow().value {
-                return nl.tight;
+        match tmp.data.borrow().value {
+            NodeValue::Item(..) | NodeValue::TaskItem(..) => {
+                if let NodeValue::List(ref nl) = tmp.parent().unwrap().data.borrow().value {
+                    return nl.tight;
+                }
+                return false;
             }
-            return false;
+            _ => {}
         }
 
         let parent = match tmp.parent() {
@@ -285,10 +302,13 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
             None => return false,
         };
 
-        if let NodeValue::Item(..) = parent.data.borrow().value {
-            if let NodeValue::List(ref nl) = parent.parent().unwrap().data.borrow().value {
-                return nl.tight;
+        match parent.data.borrow().value {
+            NodeValue::Item(..) | NodeValue::TaskItem(..) => {
+                if let NodeValue::List(ref nl) = parent.parent().unwrap().data.borrow().value {
+                    return nl.tight;
+                }
             }
+            _ => {}
         }
 
         false
@@ -298,16 +318,31 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
         self.node = node;
         let allow_wrap = self.options.render.width > 0 && !self.options.render.hardbreaks;
 
-        if !(matches!(node.data.borrow().value, NodeValue::Item(..))
-            && node.previous_sibling().is_none()
-            && entering)
-        {
-            self.in_tight_list_item = self.get_in_tight_list_item(node);
+        let parent_node = node.parent();
+        if entering {
+            if parent_node.is_some()
+                && matches!(
+                    parent_node.unwrap().data.borrow().value,
+                    NodeValue::Item(..) | NodeValue::TaskItem(..)
+                )
+            {
+                self.in_tight_list_item = self.get_in_tight_list_item(node);
+            }
+        } else if matches!(node.data.borrow().value, NodeValue::List(..)) {
+            self.in_tight_list_item = parent_node.is_some()
+                && matches!(
+                    parent_node.unwrap().data.borrow().value,
+                    NodeValue::Item(..) | NodeValue::TaskItem(..)
+                )
+                && self.get_in_tight_list_item(node);
         }
+        let next_is_block = node
+            .next_sibling()
+            .map_or(true, |next| next.data.borrow().value.block());
 
         match node.data.borrow().value {
             NodeValue::Document => (),
-            NodeValue::FrontMatter(ref fm) => self.format_front_matter(fm, entering),
+            NodeValue::FrontMatter(ref fm) => self.format_front_matter(fm.as_bytes(), entering),
             NodeValue::BlockQuote => self.format_block_quote(entering),
             NodeValue::List(..) => self.format_list(node, entering),
             NodeValue::Item(..) => self.format_item(node, entering),
@@ -320,28 +355,58 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
             NodeValue::HtmlBlock(ref nhb) => self.format_html_block(nhb, entering),
             NodeValue::ThematicBreak => self.format_thematic_break(entering),
             NodeValue::Paragraph => self.format_paragraph(entering),
-            NodeValue::Text(ref literal) => self.format_text(literal, allow_wrap, entering),
-            NodeValue::LineBreak => self.format_line_break(entering),
+            NodeValue::Text(ref literal) => {
+                self.format_text(literal.as_bytes(), allow_wrap, entering)
+            }
+            NodeValue::LineBreak => self.format_line_break(entering, next_is_block),
             NodeValue::SoftBreak => self.format_soft_break(allow_wrap, entering),
-            NodeValue::Code(ref code) => self.format_code(&code.literal, allow_wrap, entering),
-            NodeValue::HtmlInline(ref literal) => self.format_html_inline(literal, entering),
-            NodeValue::Strong => self.format_strong(),
+            NodeValue::Code(ref code) => {
+                self.format_code(code.literal.as_bytes(), allow_wrap, entering)
+            }
+            NodeValue::HtmlInline(ref literal) => {
+                self.format_html_inline(literal.as_bytes(), entering)
+            }
+            NodeValue::Raw(ref literal) => self.format_raw(literal.as_bytes(), entering),
+            NodeValue::Strong => {
+                if parent_node.is_none()
+                    || !matches!(parent_node.unwrap().data.borrow().value, NodeValue::Strong)
+                {
+                    self.format_strong();
+                }
+            }
             NodeValue::Emph => self.format_emph(node),
-            NodeValue::TaskItem(checked) => self.format_task_item(checked, entering),
+            NodeValue::TaskItem(symbol) => self.format_task_item(symbol, node, entering),
             NodeValue::Strikethrough => self.format_strikethrough(),
             NodeValue::Superscript => self.format_superscript(),
             NodeValue::Link(ref nl) => return self.format_link(node, nl, entering),
             NodeValue::Image(ref nl) => self.format_image(nl, allow_wrap, entering),
+            #[cfg(feature = "shortcodes")]
+            NodeValue::ShortCode(ref ne) => self.format_shortcode(ne, entering),
             NodeValue::Table(..) => self.format_table(entering),
             NodeValue::TableRow(..) => self.format_table_row(entering),
             NodeValue::TableCell => self.format_table_cell(node, entering),
-            NodeValue::FootnoteDefinition(_) => self.format_footnote_definition(entering),
-            NodeValue::FootnoteReference(ref r) => self.format_footnote_reference(r, entering),
+            NodeValue::FootnoteDefinition(ref nfd) => {
+                self.format_footnote_definition(&nfd.name, entering)
+            }
+            NodeValue::FootnoteReference(ref nfr) => {
+                self.format_footnote_reference(nfr.name.as_bytes(), entering)
+            }
+            NodeValue::MultilineBlockQuote(..) => self.format_block_quote(entering),
+            NodeValue::Escaped => {
+                // noop - automatic escaping is already being done
+            }
+            NodeValue::Math(ref math) => self.format_math(math, allow_wrap, entering),
+            NodeValue::WikiLink(ref nl) => return self.format_wikilink(nl, entering),
+            NodeValue::Underline => self.format_underline(),
+            NodeValue::Subscript => self.format_subscript(),
+            NodeValue::SpoileredText => self.format_spoiler(),
+            NodeValue::EscapedTag(ref net) => self.format_escaped_tag(net),
+            NodeValue::Alert(ref alert) => self.format_alert(alert, entering),
         };
         true
     }
 
-    fn format_front_matter(&mut self, front_matter: &Vec<u8>, entering: bool) {
+    fn format_front_matter(&mut self, front_matter: &[u8], entering: bool) {
         if entering {
             self.output(front_matter, false, Escaping::Literal);
         }
@@ -360,18 +425,35 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
     }
 
     fn format_list(&mut self, node: &'a AstNode<'a>, entering: bool) {
-        if !entering
-            && match node.next_sibling() {
+        let ol_start = match node.data.borrow().value {
+            NodeValue::List(NodeList {
+                list_type: ListType::Ordered,
+                start,
+                ..
+            }) => Some(start),
+            _ => None,
+        };
+
+        if entering {
+            if let Some(start) = ol_start {
+                self.ol_stack.push(start);
+            }
+        } else {
+            if ol_start.is_some() {
+                self.ol_stack.pop();
+            }
+
+            if match node.next_sibling() {
                 Some(next_sibling) => matches!(
                     next_sibling.data.borrow().value,
                     NodeValue::CodeBlock(..) | NodeValue::List(..)
                 ),
                 _ => false,
+            } {
+                self.cr();
+                write!(self, "<!-- end list -->").unwrap();
+                self.blankline();
             }
-        {
-            self.cr();
-            write!(self, "<!-- end list -->").unwrap();
-            self.blankline();
         }
     }
 
@@ -386,31 +468,45 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
         let marker_width = if parent.list_type == ListType::Bullet {
             2
         } else {
-            let mut list_number = parent.start;
+            let list_number = if let Some(last_stack) = self.ol_stack.last_mut() {
+                let list_number = *last_stack;
+                if entering {
+                    *last_stack += 1;
+                };
+                list_number
+            } else {
+                match node.data.borrow().value {
+                    NodeValue::Item(ref ni) => ni.start,
+                    NodeValue::TaskItem(_) => parent.start,
+                    _ => unreachable!(),
+                }
+            };
             let list_delim = parent.delimiter;
-            let mut tmpch = node;
-            while let Some(tmp) = tmpch.previous_sibling() {
-                tmpch = tmp;
-                list_number += 1;
-            }
             write!(
                 listmarker,
-                "{}{}{}",
+                "{}{} ",
                 list_number,
                 if list_delim == ListDelimType::Paren {
                     ")"
                 } else {
                     "."
-                },
-                if list_number < 10 { "  " } else { " " }
+                }
             )
             .unwrap();
+            let mut current_len = listmarker.len();
+
+            while current_len < self.options.render.ol_width {
+                write!(listmarker, " ").unwrap();
+                current_len += 1;
+            }
+
             listmarker.len()
         };
 
         if entering {
             if parent.list_type == ListType::Bullet {
-                write!(self, "- ").unwrap();
+                let bullet = char::from(self.options.render.list_style as u8);
+                write!(self, "{} ", bullet).unwrap();
             } else {
                 self.write_all(&listmarker).unwrap();
             }
@@ -419,7 +515,11 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
                 write!(self.prefix, " ").unwrap();
             }
         } else {
-            let new_len = self.prefix.len() - marker_width;
+            let new_len = if self.prefix.len() > marker_width {
+                self.prefix.len() - marker_width
+            } else {
+                0
+            };
             self.prefix.truncate(new_len);
             self.cr();
         }
@@ -450,7 +550,10 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
             let first_in_list_item = node.previous_sibling().is_none()
                 && match node.parent() {
                     Some(parent) => {
-                        matches!(parent.data.borrow().value, NodeValue::Item(..))
+                        matches!(
+                            parent.data.borrow().value,
+                            NodeValue::Item(..) | NodeValue::TaskItem(..)
+                        )
                     }
                     _ => false,
                 };
@@ -459,30 +562,34 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
                 self.blankline();
             }
 
-            if ncb.info.is_empty()
-                && (ncb.literal.len() > 2
-                    && !isspace(ncb.literal[0])
-                    && !(isspace(ncb.literal[ncb.literal.len() - 1])
-                        && isspace(ncb.literal[ncb.literal.len() - 2])))
-                && !first_in_list_item
+            let info = ncb.info.as_bytes();
+            let literal = ncb.literal.as_bytes();
+
+            #[allow(clippy::len_zero)]
+            if !(info.len() > 0
+                || literal.len() <= 2
+                || isspace(literal[0])
+                || first_in_list_item
+                || self.options.render.prefer_fenced
+                || isspace(literal[literal.len() - 1]) && isspace(literal[literal.len() - 2]))
             {
                 write!(self, "    ").unwrap();
                 write!(self.prefix, "    ").unwrap();
-                self.write_all(&ncb.literal).unwrap();
+                self.write_all(literal).unwrap();
                 let new_len = self.prefix.len() - 4;
                 self.prefix.truncate(new_len);
             } else {
-                let fence_char = if ncb.info.contains(&b'`') { b'~' } else { b'`' };
-                let numticks = max(3, longest_char_sequence(&ncb.literal, fence_char) + 1);
+                let fence_char = if info.contains(&b'`') { b'~' } else { b'`' };
+                let numticks = max(3, longest_char_sequence(literal, fence_char) + 1);
                 for _ in 0..numticks {
                     write!(self, "{}", fence_char as char).unwrap();
                 }
-                if !ncb.info.is_empty() {
+                if !info.is_empty() {
                     write!(self, " ").unwrap();
-                    self.write_all(&ncb.info).unwrap();
+                    self.write_all(info).unwrap();
                 }
                 self.cr();
-                self.write_all(&ncb.literal).unwrap();
+                self.write_all(literal).unwrap();
                 self.cr();
                 for _ in 0..numticks {
                     write!(self, "{}", fence_char as char).unwrap();
@@ -495,7 +602,7 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
     fn format_html_block(&mut self, nhb: &NodeHtmlBlock, entering: bool) {
         if entering {
             self.blankline();
-            self.write_all(&nhb.literal).unwrap();
+            self.write_all(nhb.literal.as_bytes()).unwrap();
             self.blankline();
         }
     }
@@ -514,16 +621,20 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
         }
     }
 
-    fn format_text(&mut self, literal: &Vec<u8>, allow_wrap: bool, entering: bool) {
+    fn format_text(&mut self, literal: &[u8], allow_wrap: bool, entering: bool) {
         if entering {
             self.output(literal, allow_wrap, Escaping::Normal);
         }
     }
 
-    fn format_line_break(&mut self, entering: bool) {
+    fn format_line_break(&mut self, entering: bool, next_is_block: bool) {
         if entering {
-            if !self.options.render.hardbreaks {
-                write!(self, "  ").unwrap();
+            if !self.options.render.hardbreaks && !next_is_block {
+                // If the next element is a block, a backslash means a
+                // literal backslash instead of a line break. In this case
+                // we can just skip the line break since it's meaningless
+                // before a block.
+                write!(self, "\\").unwrap();
             }
             self.cr();
         }
@@ -536,23 +647,28 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
                 && !self.options.render.hardbreaks
             {
                 self.cr();
+            } else if self.options.render.hardbreaks {
+                self.output(&[b'\n'], allow_wrap, Escaping::Literal);
             } else {
                 self.output(&[b' '], allow_wrap, Escaping::Literal);
             }
         }
     }
 
-    fn format_code(&mut self, literal: &Vec<u8>, allow_wrap: bool, entering: bool) {
+    fn format_code(&mut self, literal: &[u8], allow_wrap: bool, entering: bool) {
         if entering {
             let numticks = shortest_unused_sequence(literal, b'`');
             for _ in 0..numticks {
                 write!(self, "`").unwrap();
             }
-            let pad = literal.is_empty()
-                || literal[0] == b'`'
-                || literal[literal.len() - 1] == b'`'
-                || literal[0] == b' '
-                || literal[literal.len() - 1] == b' ';
+
+            let all_space = literal
+                .iter()
+                .all(|&c| c == b' ' || c == b'\r' || c == b'\n');
+            let has_edge_space = literal[0] == b' ' || literal[literal.len() - 1] == b' ';
+            let has_edge_backtick = literal[0] == b'`' || literal[literal.len() - 1] == b'`';
+
+            let pad = literal.is_empty() || has_edge_backtick || (!all_space && has_edge_space);
             if pad {
                 write!(self, " ").unwrap();
             }
@@ -566,7 +682,13 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
         }
     }
 
-    fn format_html_inline(&mut self, literal: &Vec<u8>, entering: bool) {
+    fn format_html_inline(&mut self, literal: &[u8], entering: bool) {
+        if entering {
+            self.write_all(literal).unwrap();
+        }
+    }
+
+    fn format_raw(&mut self, literal: &[u8], entering: bool) {
         if entering {
             self.write_all(literal).unwrap();
         }
@@ -591,50 +713,75 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
         self.write_all(&[emph_delim]).unwrap();
     }
 
-    fn format_task_item(&mut self, checked: bool, entering: bool) {
+    fn format_task_item(&mut self, symbol: Option<char>, node: &'a AstNode<'a>, entering: bool) {
+        self.format_item(node, entering);
         if entering {
-            if checked {
-                write!(self, "[x] ").unwrap();
-            } else {
-                write!(self, "[ ] ").unwrap();
-            }
+            write!(self, "[{}] ", symbol.unwrap_or(' ')).unwrap();
         }
     }
 
     fn format_strikethrough(&mut self) {
-        write!(self, "~").unwrap();
+        write!(self, "~~").unwrap();
     }
 
     fn format_superscript(&mut self) {
         write!(self, "^").unwrap();
     }
 
+    fn format_underline(&mut self) {
+        write!(self, "__").unwrap();
+    }
+
+    fn format_subscript(&mut self) {
+        write!(self, "~").unwrap();
+    }
+
+    fn format_spoiler(&mut self) {
+        write!(self, "||").unwrap();
+    }
+
+    fn format_escaped_tag(&mut self, net: &String) {
+        self.output(net.as_bytes(), false, Escaping::Literal);
+    }
+
     fn format_link(&mut self, node: &'a AstNode<'a>, nl: &NodeLink, entering: bool) -> bool {
         if is_autolink(node, nl) {
             if entering {
-                write!(self, "<").unwrap();
-                if nl.url.len() >= 7 && &nl.url[..7] == b"mailto:" {
-                    self.write_all(&nl.url[7..]).unwrap();
-                } else {
-                    self.write_all(&nl.url).unwrap();
-                }
-                write!(self, ">").unwrap();
+                write!(self, "<{}>", trim_start_match(&nl.url, "mailto:")).unwrap();
                 return false;
             }
         } else if entering {
             write!(self, "[").unwrap();
         } else {
             write!(self, "](").unwrap();
-            self.output(&nl.url, false, Escaping::URL);
+            self.output(nl.url.as_bytes(), false, Escaping::Url);
             if !nl.title.is_empty() {
                 write!(self, " \"").unwrap();
-                self.output(&nl.title, false, Escaping::Title);
+                self.output(nl.title.as_bytes(), false, Escaping::Title);
                 write!(self, "\"").unwrap();
             }
             write!(self, ")").unwrap();
         }
 
-        return true;
+        true
+    }
+
+    fn format_wikilink(&mut self, nl: &NodeWikiLink, entering: bool) -> bool {
+        if entering {
+            write!(self, "[[").unwrap();
+            if self.options.extension.wikilinks() == Some(WikiLinksMode::UrlFirst) {
+                self.output(nl.url.as_bytes(), false, Escaping::Url);
+                write!(self, "|").unwrap();
+            }
+        } else {
+            if self.options.extension.wikilinks() == Some(WikiLinksMode::TitleFirst) {
+                write!(self, "|").unwrap();
+                self.output(nl.url.as_bytes(), false, Escaping::Url);
+            }
+            write!(self, "]]").unwrap();
+        }
+
+        true
     }
 
     fn format_image(&mut self, nl: &NodeLink, allow_wrap: bool, entering: bool) {
@@ -642,16 +789,25 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
             write!(self, "![").unwrap();
         } else {
             write!(self, "](").unwrap();
-            self.output(&nl.url, false, Escaping::URL);
+            self.output(nl.url.as_bytes(), false, Escaping::Url);
             if !nl.title.is_empty() {
                 self.output(&[b' ', b'"'], allow_wrap, Escaping::Literal);
-                self.output(&nl.title, false, Escaping::Title);
+                self.output(nl.title.as_bytes(), false, Escaping::Title);
                 write!(self, "\"").unwrap();
             }
             write!(self, ")").unwrap();
         }
     }
 
+    #[cfg(feature = "shortcodes")]
+    fn format_shortcode(&mut self, ne: &NodeShortCode, entering: bool) {
+        if entering {
+            write!(self, ":").unwrap();
+            self.output(ne.code.as_bytes(), false, Escaping::Literal);
+            write!(self, ":").unwrap();
+        }
+    }
+
     fn format_table(&mut self, entering: bool) {
         if entering {
             self.custom_escape = Some(table_escape);
@@ -683,7 +839,7 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
             if in_header && node.next_sibling().is_none() {
                 let table = &node.parent().unwrap().parent().unwrap().data.borrow().value;
                 let alignments = match *table {
-                    NodeValue::Table(ref alignments) => alignments,
+                    NodeValue::Table(NodeTable { ref alignments, .. }) => alignments,
                     _ => panic!(),
                 };
 
@@ -706,11 +862,10 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
             }
         }
     }
-    fn format_footnote_definition(&mut self, entering: bool) {
+    fn format_footnote_definition(&mut self, name: &str, entering: bool) {
         if entering {
             self.footnote_ix += 1;
-            let footnote_ix = self.footnote_ix;
-            writeln!(self, "[^{}]:", footnote_ix).unwrap();
+            writeln!(self, "[^{}]:", name).unwrap();
             write!(self.prefix, "    ").unwrap();
         } else {
             let new_len = self.prefix.len() - 4;
@@ -718,13 +873,61 @@ impl<'a, 'o> CommonMarkFormatter<'a, 'o> {
         }
     }
 
-    fn format_footnote_reference(&mut self, r: &Vec<u8>, entering: bool) {
+    fn format_footnote_reference(&mut self, r: &[u8], entering: bool) {
         if entering {
             self.write_all(b"[^").unwrap();
             self.write_all(r).unwrap();
             self.write_all(b"]").unwrap();
         }
     }
+
+    fn format_math(&mut self, math: &NodeMath, allow_wrap: bool, entering: bool) {
+        if entering {
+            let literal = math.literal.as_bytes();
+            let start_fence = if math.dollar_math {
+                if math.display_math {
+                    "$$"
+                } else {
+                    "$"
+                }
+            } else {
+                "$`"
+            };
+
+            let end_fence = if start_fence == "$`" {
+                "`$"
+            } else {
+                start_fence
+            };
+
+            self.output(start_fence.as_bytes(), false, Escaping::Literal);
+            self.output(literal, allow_wrap, Escaping::Literal);
+            self.output(end_fence.as_bytes(), false, Escaping::Literal);
+        }
+    }
+
+    fn format_alert(&mut self, alert: &NodeAlert, entering: bool) {
+        if entering {
+            write!(
+                self,
+                "> [!{}]",
+                alert.alert_type.default_title().to_uppercase()
+            )
+            .unwrap();
+            if alert.title.is_some() {
+                let title = alert.title.as_ref().unwrap();
+                write!(self, " {}", title).unwrap();
+            }
+            writeln!(self).unwrap();
+            write!(self, "> ").unwrap();
+            self.begin_content = true;
+            write!(self.prefix, "> ").unwrap();
+        } else {
+            let new_len = self.prefix.len() - 2;
+            self.prefix.truncate(new_len);
+            self.blankline();
+        }
+    }
 }
 
 fn longest_char_sequence(literal: &[u8], ch: u8) -> usize {
@@ -773,7 +976,7 @@ fn shortest_unused_sequence(literal: &[u8], f: u8) -> usize {
 }
 
 fn is_autolink<'a>(node: &'a AstNode<'a>, nl: &NodeLink) -> bool {
-    if nl.url.is_empty() || scanners::scheme(&nl.url).is_none() {
+    if nl.url.is_empty() || scanners::scheme(nl.url.as_bytes()).is_none() {
         return false;
     }
 
@@ -789,12 +992,7 @@ fn is_autolink<'a>(node: &'a AstNode<'a>, nl: &NodeLink) -> bool {
         },
     };
 
-    let mut real_url: &[u8] = &nl.url;
-    if real_url.len() >= 7 && &real_url[..7] == b"mailto:" {
-        real_url = &real_url[7..];
-    }
-
-    real_url == &*link_text
+    trim_start_match(&nl.url, "mailto:") == link_text
 }
 
 fn table_escape<'a>(node: &'a AstNode<'a>, c: u8) -> bool {
