@@ -1,42 +1,51 @@
-use ctype::{isalnum, isalpha, isspace};
-use nodes::{AstNode, NodeLink, NodeValue};
-use parser::inlines::make_inline;
+use crate::character_set::character_set;
+use crate::ctype::{isalnum, isalpha, isspace};
+use crate::nodes::{AstNode, NodeLink, NodeValue};
+use crate::parser::inlines::make_inline;
 use std::str;
 use typed_arena::Arena;
 use unicode_categories::UnicodeCategories;
 
-pub fn process_autolinks<'a>(
+// TODO: this can probably be cleaned up a lot. It used to handle all three of
+// {url,www,email}_match, but now just the last of those.
+pub(crate) fn process_autolinks<'a>(
     arena: &'a Arena<AstNode<'a>>,
     node: &'a AstNode<'a>,
-    contents: &mut Vec<u8>,
+    contents_str: &mut String,
+    relaxed_autolinks: bool,
 ) {
+    let contents = contents_str.as_bytes();
     let len = contents.len();
     let mut i = 0;
 
     while i < len {
         let mut post_org = None;
+        let mut bracket_opening = 0;
 
+        // cmark-gfm ignores links inside brackets, such as `[[http://example.com]`
         while i < len {
-            match contents[i] {
-                b':' => {
-                    post_org = url_match(arena, contents, i);
-                    if post_org.is_some() {
-                        break;
+            if !relaxed_autolinks {
+                match contents[i] {
+                    b'[' => {
+                        bracket_opening += 1;
                     }
-                }
-                b'w' => {
-                    post_org = www_match(arena, contents, i);
-                    if post_org.is_some() {
-                        break;
+                    b']' => {
+                        bracket_opening -= 1;
                     }
+                    _ => (),
                 }
-                b'@' => {
-                    post_org = email_match(arena, contents, i);
-                    if post_org.is_some() {
-                        break;
-                    }
+
+                if bracket_opening > 0 {
+                    i += 1;
+                    continue;
+                }
+            }
+
+            if contents[i] == b'@' {
+                post_org = email_match(arena, contents, i, relaxed_autolinks);
+                if post_org.is_some() {
+                    break;
                 }
-                _ => (),
             }
             i += 1;
         }
@@ -45,30 +54,27 @@ pub fn process_autolinks<'a>(
             i -= reverse;
             node.insert_after(post);
             if i + skip < len {
-                let remain = contents[i + skip..].to_vec();
+                let remain = str::from_utf8(&contents[i + skip..]).unwrap();
                 assert!(!remain.is_empty());
-                post.insert_after(make_inline(arena, NodeValue::Text(remain)));
+                post.insert_after(make_inline(
+                    arena,
+                    NodeValue::Text(remain.to_string()),
+                    (0, 1, 0, 1).into(),
+                ));
             }
-            contents.truncate(i);
+            contents_str.truncate(i);
             return;
         }
     }
 }
 
-fn www_match<'a>(
+pub fn www_match<'a>(
     arena: &'a Arena<AstNode<'a>>,
     contents: &[u8],
     i: usize,
+    relaxed_autolinks: bool,
 ) -> Option<(&'a AstNode<'a>, usize, usize)> {
-    lazy_static! {
-        static ref WWW_DELIMS: [bool; 256] = {
-            let mut sc = [false; 256];
-            for c in &[b'*', b'_', b'~', b'(', b'['] {
-                sc[*c as usize] = true;
-            }
-            sc
-        };
-    }
+    const WWW_DELIMS: [bool; 256] = character_set!(b"*_~([");
 
     if i > 0 && !isspace(contents[i - 1]) && !WWW_DELIMS[contents[i - 1] as usize] {
         return None;
@@ -84,19 +90,36 @@ fn www_match<'a>(
     };
 
     while i + link_end < contents.len() && !isspace(contents[i + link_end]) {
+        // basic test to detect whether we're in a normal markdown link - not exhaustive
+        if relaxed_autolinks && contents[i + link_end - 1] == b']' && contents[i + link_end] == b'('
+        {
+            return None;
+        }
         link_end += 1;
     }
 
-    link_end = autolink_delim(&contents[i..], link_end);
+    link_end = autolink_delim(&contents[i..], link_end, relaxed_autolinks);
 
-    let mut url = b"http://".to_vec();
-    url.extend_from_slice(&contents[i..link_end + i]);
+    let mut url = "http://".to_string();
+    url.push_str(str::from_utf8(&contents[i..link_end + i]).unwrap());
 
-    let inl = make_inline(arena, NodeValue::Link(NodeLink { url, title: vec![] }));
+    let inl = make_inline(
+        arena,
+        NodeValue::Link(NodeLink {
+            url,
+            title: String::new(),
+        }),
+        (0, 1, 0, 1).into(),
+    );
 
     inl.append(make_inline(
         arena,
-        NodeValue::Text(contents[i..link_end + i].to_vec()),
+        NodeValue::Text(
+            str::from_utf8(&contents[i..link_end + i])
+                .unwrap()
+                .to_string(),
+        ),
+        (0, 1, 0, 1).into(),
     ));
     Some((inl, 0, link_end))
 }
@@ -107,7 +130,10 @@ fn check_domain(data: &[u8], allow_short: bool) -> Option<usize> {
     let mut uscore2 = 0;
 
     for (i, c) in unsafe { str::from_utf8_unchecked(data) }.char_indices() {
-        if c == '_' {
+        if c == '\\' && i < data.len() - 1 {
+            // Ignore escaped characters per https://github.com/github/cmark-gfm/pull/292.
+            // Not sure I love this, but it tracks upstream ..
+        } else if c == '_' {
             uscore2 += 1;
         } else if c == '.' {
             uscore1 = uscore2;
@@ -121,7 +147,7 @@ fn check_domain(data: &[u8], allow_short: bool) -> Option<usize> {
         }
     }
 
-    if uscore1 > 0 || uscore2 > 0 {
+    if (uscore1 > 0 || uscore2 > 0) && np <= 10 {
         None
     } else if allow_short || np > 0 {
         Some(data.len())
@@ -131,19 +157,11 @@ fn check_domain(data: &[u8], allow_short: bool) -> Option<usize> {
 }
 
 fn is_valid_hostchar(ch: char) -> bool {
-    !ch.is_whitespace() && !ch.is_punctuation()
+    !(ch.is_whitespace() || ch.is_punctuation() || ch.is_symbol())
 }
 
-fn autolink_delim(data: &[u8], mut link_end: usize) -> usize {
-    lazy_static! {
-        static ref LINK_END_ASSORTMENT: [bool; 256] = {
-            let mut sc = [false; 256];
-            for c in &[b'?', b'!', b'.', b',', b':', b'*', b'_', b'~', b'\'', b'"'] {
-                sc[*c as usize] = true;
-            }
-            sc
-        };
-    }
+fn autolink_delim(data: &[u8], mut link_end: usize, relaxed_autolinks: bool) -> usize {
+    const LINK_END_ASSORTMENT: [bool; 256] = character_set!(b"?!.,:*_~'\"");
 
     for (i, &b) in data.iter().enumerate().take(link_end) {
         if b == b'<' {
@@ -155,7 +173,22 @@ fn autolink_delim(data: &[u8], mut link_end: usize) -> usize {
     while link_end > 0 {
         let cclose = data[link_end - 1];
 
-        let copen = if cclose == b')' { Some(b'(') } else { None };
+        // Allow any number of matching parentheses (as recognised in copen/cclose)
+        // at the end of the URL.  If there is a greater number of closing
+        // parentheses than opening ones, we remove one character from the end of
+        // the link.
+        let mut copen = if cclose == b')' { Some(b'(') } else { None };
+
+        if relaxed_autolinks && copen.is_none() {
+            // allow balancing of `[]` and `{}` just like `()`
+            copen = if cclose == b']' {
+                Some(b'[')
+            } else if cclose == b'}' {
+                Some(b'{')
+            } else {
+                None
+            };
+        }
 
         if LINK_END_ASSORTMENT[cclose as usize] {
             link_end -= 1;
@@ -195,14 +228,13 @@ fn autolink_delim(data: &[u8], mut link_end: usize) -> usize {
     link_end
 }
 
-fn url_match<'a>(
+pub fn url_match<'a>(
     arena: &'a Arena<AstNode<'a>>,
     contents: &[u8],
     i: usize,
+    relaxed_autolinks: bool,
 ) -> Option<(&'a AstNode<'a>, usize, usize)> {
-    lazy_static! {
-        static ref SCHEMES: Vec<&'static [u8]> = vec![b"http", b"https", b"ftp"];
-    }
+    const SCHEMES: [&[u8]; 3] = [b"http", b"https", b"ftp"];
 
     let size = contents.len();
 
@@ -215,9 +247,11 @@ fn url_match<'a>(
         rewind += 1;
     }
 
-    let cond = |s: &&[u8]| size - i + rewind >= s.len() && &&contents[i - rewind..i] == s;
-    if !SCHEMES.iter().any(cond) {
-        return None;
+    if !relaxed_autolinks {
+        let cond = |s: &&[u8]| size - i + rewind >= s.len() && &&contents[i - rewind..i] == s;
+        if !SCHEMES.iter().any(cond) {
+            return None;
+        }
     }
 
     let mut link_end = match check_domain(&contents[i + 3..], true) {
@@ -226,21 +260,36 @@ fn url_match<'a>(
     };
 
     while link_end < size - i && !isspace(contents[i + link_end]) {
+        // basic test to detect whether we're in a normal markdown link - not exhaustive
+        if relaxed_autolinks
+            && link_end > 0
+            && contents[i + link_end - 1] == b']'
+            && contents[i + link_end] == b'('
+        {
+            return None;
+        }
         link_end += 1;
     }
 
-    link_end = autolink_delim(&contents[i..], link_end);
+    link_end = autolink_delim(&contents[i..], link_end, relaxed_autolinks);
 
-    let url = contents[i - rewind..i + link_end].to_vec();
+    let url = str::from_utf8(&contents[i - rewind..i + link_end])
+        .unwrap()
+        .to_string();
     let inl = make_inline(
         arena,
         NodeValue::Link(NodeLink {
             url: url.clone(),
-            title: vec![],
+            title: String::new(),
         }),
+        (0, 1, 0, 1).into(),
     );
 
-    inl.append(make_inline(arena, NodeValue::Text(url)));
+    inl.append(make_inline(
+        arena,
+        NodeValue::Text(url),
+        (0, 1, 0, 1).into(),
+    ));
     Some((inl, rewind, rewind + link_end))
 }
 
@@ -248,21 +297,15 @@ fn email_match<'a>(
     arena: &'a Arena<AstNode<'a>>,
     contents: &[u8],
     i: usize,
+    relaxed_autolinks: bool,
 ) -> Option<(&'a AstNode<'a>, usize, usize)> {
-    lazy_static! {
-        static ref EMAIL_OK_SET: [bool; 256] = {
-            let mut sc = [false; 256];
-            for c in &[b'.', b'+', b'-', b'_'] {
-                sc[*c as usize] = true;
-            }
-            sc
-        };
-    }
+    const EMAIL_OK_SET: [bool; 256] = character_set!(b".+-_");
 
     let size = contents.len();
 
+    let mut auto_mailto = true;
+    let mut is_xmpp = false;
     let mut rewind = 0;
-    let mut ns = 0;
 
     while rewind < i {
         let c = contents[i - rewind - 1];
@@ -272,19 +315,29 @@ fn email_match<'a>(
             continue;
         }
 
-        if c == b'/' {
-            ns += 1;
+        if c == b':' {
+            if validate_protocol("mailto", contents, i - rewind - 1) {
+                auto_mailto = false;
+                rewind += 1;
+                continue;
+            }
+
+            if validate_protocol("xmpp", contents, i - rewind - 1) {
+                is_xmpp = true;
+                auto_mailto = false;
+                rewind += 1;
+                continue;
+            }
         }
 
         break;
     }
 
-    if rewind == 0 || ns > 0 {
+    if rewind == 0 {
         return None;
     }
 
-    let mut link_end = 0;
-    let mut nb = 0;
+    let mut link_end = 1;
     let mut np = 0;
 
     while link_end < size - i {
@@ -293,9 +346,11 @@ fn email_match<'a>(
         if isalnum(c) {
             // empty
         } else if c == b'@' {
-            nb += 1;
+            return None;
         } else if c == b'.' && link_end < size - i - 1 && isalnum(contents[i + link_end + 1]) {
             np += 1;
+        } else if c == b'/' && is_xmpp {
+            // xmpp allows a `/` in the url
         } else if c != b'-' && c != b'_' {
             break;
         }
@@ -304,23 +359,50 @@ fn email_match<'a>(
     }
 
     if link_end < 2
-        || nb != 1
         || np == 0
         || (!isalpha(contents[i + link_end - 1]) && contents[i + link_end - 1] != b'.')
     {
         return None;
     }
 
-    link_end = autolink_delim(&contents[i..], link_end);
+    link_end = autolink_delim(&contents[i..], link_end, relaxed_autolinks);
+    if link_end == 0 {
+        return None;
+    }
 
-    let mut url = b"mailto:".to_vec();
-    url.extend_from_slice(&contents[i - rewind..link_end + i]);
+    let mut url = if auto_mailto {
+        "mailto:".to_string()
+    } else {
+        "".to_string()
+    };
+    let text = str::from_utf8(&contents[i - rewind..link_end + i]).unwrap();
+    url.push_str(text);
 
-    let inl = make_inline(arena, NodeValue::Link(NodeLink { url, title: vec![] }));
+    let inl = make_inline(
+        arena,
+        NodeValue::Link(NodeLink {
+            url,
+            title: String::new(),
+        }),
+        (0, 1, 0, 1).into(),
+    );
 
     inl.append(make_inline(
         arena,
-        NodeValue::Text(contents[i - rewind..link_end + i].to_vec()),
+        NodeValue::Text(text.to_string()),
+        (0, 1, 0, 1).into(),
     ));
     Some((inl, rewind, rewind + link_end))
 }
+
+fn validate_protocol(protocol: &str, contents: &[u8], cursor: usize) -> bool {
+    let size = contents.len();
+    let mut rewind = 0;
+
+    while rewind < cursor && isalpha(contents[cursor - rewind - 1]) {
+        rewind += 1;
+    }
+
+    size - cursor + rewind >= protocol.len()
+        && &contents[cursor - rewind..cursor] == protocol.as_bytes()
+}
