@@ -1,10 +1,15 @@
-use ctype::{ispunct, isspace};
-use entity;
-use parser::AutolinkType;
-use std::collections::HashMap;
+use crate::ctype::{ispunct, isspace};
+use crate::entity;
+use crate::parser::AutolinkType;
 use std::ptr;
 use std::str;
 
+#[derive(PartialEq, Eq)]
+pub enum Case {
+    Preserve,
+    Fold,
+}
+
 pub fn unescape(v: &mut Vec<u8>) {
     let mut r = 0;
     let mut prev = None;
@@ -12,6 +17,10 @@ pub fn unescape(v: &mut Vec<u8>) {
 
     while r < v.len() {
         if v[r] == b'\\' && r + 1 < v.len() && ispunct(v[r + 1]) {
+            if v[r + 1] == b'\\' {
+                r += 1;
+            }
+
             if let Some(prev) = prev {
                 let window = &mut v[(prev + 1 - found)..r];
                 shift_buf_left(window, found);
@@ -65,7 +74,7 @@ pub fn normalize_code(v: &[u8]) -> Vec<u8> {
             }
             c => r.push(c),
         }
-        if v[i] != b' ' {
+        if v[i] != b' ' && v[i] != b'\r' && v[i] != b'\n' {
             contains_nonspace = true;
         }
 
@@ -80,10 +89,11 @@ pub fn normalize_code(v: &[u8]) -> Vec<u8> {
     r
 }
 
-pub fn remove_trailing_blank_lines(line: &mut Vec<u8>) {
+pub fn remove_trailing_blank_lines(line: &mut String) {
+    let line_bytes = line.as_bytes();
     let mut i = line.len() - 1;
     loop {
-        let c = line[i];
+        let c = line_bytes[i];
 
         if c != b' ' && c != b'\t' && !is_line_end_char(c) {
             break;
@@ -97,10 +107,8 @@ pub fn remove_trailing_blank_lines(line: &mut Vec<u8>) {
         i -= 1;
     }
 
-    for i in i..line.len() {
-        let c = line[i];
-
-        if !is_line_end_char(c) {
+    for (i, c) in line_bytes.iter().enumerate().take(line.len()).skip(i) {
+        if !is_line_end_char(*c) {
             continue;
         }
 
@@ -154,22 +162,31 @@ pub fn trim(line: &mut Vec<u8>) {
     rtrim(line);
 }
 
+pub fn ltrim_slice(mut i: &[u8]) -> &[u8] {
+    while let [first, rest @ ..] = i {
+        if isspace(*first) {
+            i = rest;
+        } else {
+            break;
+        }
+    }
+    i
+}
+
 pub fn rtrim_slice(mut i: &[u8]) -> &[u8] {
-    let mut len = i.len();
-    while len > 0 && isspace(i[len - 1]) {
-        i = &i[..len - 1];
-        len -= 1;
+    while let [rest @ .., last] = i {
+        if isspace(*last) {
+            i = rest;
+        } else {
+            break;
+        }
     }
     i
 }
 
 pub fn trim_slice(mut i: &[u8]) -> &[u8] {
+    i = ltrim_slice(i);
     i = rtrim_slice(i);
-    let mut len = i.len();
-    while len > 0 && isspace(i[0]) {
-        i = &i[1..];
-        len -= 1;
-    }
     i
 }
 
@@ -230,47 +247,140 @@ pub fn is_blank(s: &[u8]) -> bool {
     true
 }
 
-pub fn normalize_label(i: &[u8]) -> Vec<u8> {
-    let i = trim_slice(i);
+pub fn normalize_label(i: &str, casing: Case) -> String {
+    // trim_slice only removes bytes from start and end that match isspace();
+    // result is UTF-8.
+    let i = unsafe { str::from_utf8_unchecked(trim_slice(i.as_bytes())) };
+
     let mut v = String::with_capacity(i.len());
     let mut last_was_whitespace = false;
-    for c in unsafe { str::from_utf8_unchecked(i) }.chars() {
-        for e in c.to_lowercase() {
-            if e.is_whitespace() {
-                if !last_was_whitespace {
-                    last_was_whitespace = true;
-                    v.push(' ');
-                }
-            } else {
-                last_was_whitespace = false;
-                v.push(e);
+    for c in i.chars() {
+        if c.is_whitespace() {
+            if !last_was_whitespace {
+                last_was_whitespace = true;
+                v.push(' ');
             }
+        } else {
+            last_was_whitespace = false;
+            v.push(c);
         }
     }
-    v.into_bytes()
+
+    if casing == Case::Fold {
+        caseless::default_case_fold_str(&v)
+    } else {
+        v
+    }
 }
 
-pub fn build_opening_tag(tag: &str, attributes: &HashMap<String, String>) -> String {
-    let mut tag_parts = vec![format!("<{}", tag)];
+#[test]
+fn normalize_label_fold_test() {
+    assert_eq!(normalize_label("Abc   \t\ndef", Case::Preserve), "Abc def");
+    assert_eq!(normalize_label("Abc   \t\ndef", Case::Fold), "abc def");
+    assert_eq!(normalize_label("Straẞe", Case::Preserve), "Straẞe");
+    assert_eq!(normalize_label("Straẞe", Case::Fold), "strasse");
+}
 
-    for (attr, val) in attributes {
-        tag_parts.push(format!(" {}=\"{}\"", attr, val));
+pub fn split_off_front_matter<'s>(mut s: &'s str, delimiter: &str) -> Option<(&'s str, &'s str)> {
+    s = trim_start_match(s, "\u{feff}");
+
+    if !s.starts_with(delimiter) {
+        return None;
+    }
+    let mut start = delimiter.len();
+    if s[start..].starts_with('\n') {
+        start += 1;
+    } else if s[start..].starts_with("\r\n") {
+        start += 2;
+    } else {
+        return None;
     }
 
-    tag_parts.push(String::from(">"));
+    start += match s[start..]
+        .find(&("\n".to_string() + delimiter + "\r\n"))
+        .or_else(|| s[start..].find(&("\n".to_string() + delimiter + "\n")))
+        .or_else(|| s[start..].find(&("\n".to_string() + delimiter))) // delimiter followed by EOF
+    {
+        Some(n) => n + 1 + delimiter.len(),
+        None => return None,
+    };
 
-    tag_parts.join("")
-}
-
-#[cfg(feature = "syntect")]
-pub fn extract_attributes_from_tag(html_tag: &str) -> HashMap<String, String> {
-    let re = regex::Regex::new("([a-zA-Z_:][-a-zA-Z0-9_:.]+)=([\"'])(.*?)([\"'])").unwrap();
-
-    let mut attributes: HashMap<String, String> = HashMap::new();
-
-    for caps in re.captures_iter(html_tag) {
-        attributes.insert(String::from(&caps[1]), String::from(&caps[3]));
+    if start == s.len() {
+        return Some((s, ""));
     }
 
-    attributes
+    start += if s[start..].starts_with('\n') {
+        1
+    } else if s[start..].starts_with("\r\n") {
+        2
+    } else {
+        return None;
+    };
+
+    start += if s[start..].starts_with('\n') {
+        1
+    } else if s[start..].starts_with("\r\n") {
+        2
+    } else {
+        0
+    };
+
+    Some((&s[..start], &s[start..]))
+}
+
+pub fn trim_start_match<'s>(s: &'s str, pat: &str) -> &'s str {
+    s.strip_prefix(pat).unwrap_or(s)
+}
+
+#[cfg(test)]
+pub mod tests {
+    use super::{normalize_code, normalize_label, split_off_front_matter};
+    use crate::strings::Case;
+
+    #[test]
+    fn normalize_code_handles_lone_newline() {
+        assert_eq!(normalize_code(&[b'\n']), vec![b' ']);
+    }
+
+    #[test]
+    fn normalize_code_handles_lone_space() {
+        assert_eq!(normalize_code(&[b' ']), vec![b' ']);
+    }
+
+    #[test]
+    fn front_matter() {
+        assert_eq!(
+            split_off_front_matter("---\nfoo: bar\n---\nHiiii", "---"),
+            Some(("---\nfoo: bar\n---\n", "Hiiii"))
+        );
+        assert_eq!(
+            split_off_front_matter(
+                "\u{feff}!@#\r\n\r\nfoo: !@# \r\nquux\n!@#\r\n\n\nYes!\n",
+                "!@#"
+            ),
+            Some(("!@#\r\n\r\nfoo: !@# \r\nquux\n!@#\r\n\n", "\nYes!\n"))
+        );
+        assert_eq!(
+            split_off_front_matter(
+                "\u{feff}!@#\r\n\r\nfoo: \n!@# \r\nquux\n!@#\r\n\n\nYes!\n",
+                "!@#"
+            ),
+            Some(("!@#\r\n\r\nfoo: \n!@# \r\nquux\n!@#\r\n\n", "\nYes!\n"))
+        );
+    }
+
+    #[test]
+    fn normalize_label_lowercase() {
+        assert_eq!(normalize_label("  Foo\u{A0}BAR  ", Case::Fold), "foo bar");
+        assert_eq!(normalize_label("  FooİBAR  ", Case::Fold), "fooi\u{307}bar");
+    }
+
+    #[test]
+    fn normalize_label_preserve() {
+        assert_eq!(
+            normalize_label("  Foo\u{A0}BAR  ", Case::Preserve),
+            "Foo BAR"
+        );
+        assert_eq!(normalize_label("  FooİBAR  ", Case::Preserve), "FooİBAR");
+    }
 }
